@@ -11,20 +11,40 @@ from __future__ import absolute_import, division, unicode_literals
 
 import sys
 
-from mo_future import binary_type, generator_types, is_binary, is_text, text, OrderedDict
+from mo_future import (
+    binary_type,
+    generator_types,
+    is_binary,
+    is_text,
+    text,
+    OrderedDict,
+    none_type,
+)
+from mo_imports import export
 
+from mo_dots.datas import Data, SLOT, data_types, is_data
+from mo_dots.lists import (
+    FlatList,
+    is_list,
+    is_sequence,
+    is_container,
+    is_many,
+    LIST,
+    list_types,
+    container_types, finite_types,
+)
+from mo_dots.nones import Null, NullType
+from mo_dots.objects import DataObject
 from mo_dots.utils import CLASS, OBJ, get_logger, get_module
 
-none_type = type(None)
-ModuleType = type(sys.modules[__name__])
-
-
+_module_type = type(sys.modules[__name__])
 _builtin_zip = zip
-ROOT_PATH = ["."]
-
-
 _get = object.__getattribute__
 _set = object.__setattr__
+_new = object.__new__
+
+
+ROOT_PATH = ["."]
 
 
 def inverse(d):
@@ -32,7 +52,7 @@ def inverse(d):
     reverse the k:v pairs
     """
     output = {}
-    for k, v in unwrap(d).items():
+    for k, v in from_data(d).items():
         output[v] = output.get(v, [])
         output[v].append(k)
     return output
@@ -43,7 +63,7 @@ def coalesce(*args):
     # http://en.wikipedia.org/wiki/Null_coalescing_operator
     for a in args:
         if a != None:
-            return wrap(a)
+            return to_data(a)
     return Null
 
 
@@ -61,11 +81,11 @@ def zip(keys, values):
 
 
 def missing(value):
-    return value == None or value == ''
+    return value == None or value == ""
 
 
 def exists(value):
-    return value != None and value != ''
+    return value != None and value != ""
 
 
 def literal_field(field):
@@ -98,7 +118,7 @@ def tail_field(field):
     RETURN THE FIRST STEP IN PATH, ALONG WITH THE REMAINING TAIL
     IN (first, rest) PAIR
     """
-    if field == "." or field==None:
+    if field == "." or field == None:
         return ".", "."
     elif "." in field:
         if "\\." in field:
@@ -113,18 +133,19 @@ def tail_field(field):
         return field, "."
 
 
-
 def split_field(field):
     """
     RETURN field AS ARRAY OF DOT-SEPARATED FIELDS
     """
-    if field == "." or field==None:
+    if field == "." or field == None:
         return []
     elif is_text(field) and "." in field:
         if field.startswith(".."):
             remainder = field.lstrip(".")
             back = len(field) - len(remainder) - 1
-            return [-1]*back + [k.replace("\a", ".") for k in remainder.replace("\\.", "\a").split(".")]
+            return [-1] * back + [
+                k.replace("\a", ".") for k in remainder.replace("\\.", "\a").split(".")
+            ]
         else:
             return [k.replace("\a", ".") for k in field.replace("\\.", "\a").split(".")]
     else:
@@ -135,6 +156,18 @@ def join_field(path):
     """
     RETURN field SEQUENCE AS STRING
     """
+    if path.__class__ in generator_types:
+        path = list(path)
+
+    if not path:
+        return "."
+
+    if path[0] == -1:
+        for i, step in enumerate(path):
+            if step != -1:
+                parents = "." + ("." * i)
+                return parents + ".".join([f.replace(".", "\\.") for f in path[i:] if f != None])
+        return "." + ("." * len(path))
     output = ".".join([f.replace(".", "\\.") for f in path if f != None])
     return output if output else "."
 
@@ -148,9 +181,9 @@ def concat_field(prefix, suffix):
     if suffix.startswith(".."):
         remainder = suffix.lstrip(".")
         back = len(suffix) - len(remainder) - 1
-        prefix_path=split_field(prefix)
-        if len(prefix_path)>=back:
-            return join_field(split_field(prefix)[:-back]+split_field(remainder))
+        prefix_path = split_field(prefix)
+        if len(prefix_path) >= back:
+            return join_field(split_field(prefix)[:-back] + split_field(remainder))
         else:
             return "." * (back - len(prefix_path)) + "." + remainder
     else:
@@ -182,7 +215,7 @@ def relative_field(field, parent):
     """
     RETURN field PATH WITH RESPECT TO parent
     """
-    if parent==".":
+    if parent == ".":
         return field
 
     field_path = split_field(field)
@@ -222,11 +255,11 @@ def set_default(*dicts):
     p0 = dicts[0]
     agg = p0 if p0 or _get(p0, CLASS) in data_types else {}
     for p in dicts[1:]:
-        p = unwrap(p)
+        p = from_data(p)
         if p is None:
             continue
         _all_default(agg, p, seen={})
-    return wrap(agg)
+    return to_data(agg)
 
 
 def _all_default(d, default, seen=None):
@@ -237,12 +270,12 @@ def _all_default(d, default, seen=None):
     if default is None:
         return
     if _get(default, CLASS) is Data:
-        default = object.__getattribute__(default, SLOT)  # REACH IN AND GET THE dict
+        default = _get(default, SLOT)  # REACH IN AND GET THE dict
         # Log = _late_import()
         # Log.error("strictly dict (or object) allowed: got {{type}}", type=_get(default, CLASS).__name__)
 
     for k, default_value in default.items():
-        default_value = unwrap(default_value)  # TWO DIFFERENT Dicts CAN SHARE id() BECAUSE THEY ARE SHORT LIVED
+        default_value = from_data(default_value)  # TWO DIFFERENT Dicts CAN SHARE id() BECAUSE THEY ARE SHORT LIVED
         if is_data(d):
             existing_value = d.get(k)
         else:
@@ -265,11 +298,16 @@ def _all_default(d, default, seen=None):
                         _set_attr(d, [k], default_value)
                     except Exception as e:
                         if PATH_NOT_FOUND not in e:
-                            get_logger().error("Can not set attribute {{name}}", name=k, cause=e)
+                            get_logger().error(
+                                "Can not set attribute {{name}}", name=k, cause=e
+                            )
         elif is_list(existing_value) or is_list(default_value):
             _set_attr(d, [k], None)
             _set_attr(d, [k], listwrap(existing_value) + listwrap(default_value))
-        elif (hasattr(existing_value, "__setattr__") or _get(existing_value, CLASS) in data_types) and _get(default_value, CLASS) in data_types:
+        elif (
+            hasattr(existing_value, "__setattr__")
+            or _get(existing_value, CLASS) in data_types
+        ) and _get(default_value, CLASS) in data_types:
             df = seen.get(id(default_value))
             if df is not None:
                 _set_attr(d, [k], df)
@@ -309,20 +347,21 @@ def _getdefault(obj, key):
     except Exception as f:
         pass
 
+    if is_sequence(obj):
+        return [_getdefault(o, key) for o in obj]
+
     try:
         if obj.__class__ is not dict:
             return getattr(obj, key)
     except Exception as f:
         pass
 
-
     try:
         if float(key) == round(float(key), 0):
             return obj[int(key)]
     except Exception as f:
         pass
 
-
     # TODO: FIGURE OUT WHY THIS WAS EVER HERE (AND MAKE A TEST)
     # try:
     #     return eval("obj."+text(key))
@@ -359,7 +398,7 @@ def get_attr(obj, path):
     except Exception as e:
         Log = get_logger()
         if PATH_NOT_FOUND in e:
-            Log.error(PATH_NOT_FOUND+": {{path}}",  path=path, cause=e)
+            Log.error(PATH_NOT_FOUND + ": {{path}}", path=path, cause=e)
         else:
             Log.error("Problem setting value", e)
 
@@ -370,7 +409,7 @@ def _get_attr(obj, path):
 
     attr_name = path[0]
 
-    if isinstance(obj, ModuleType):
+    if isinstance(obj, _module_type):
         if attr_name in obj.__dict__:
             return _get_attr(obj.__dict__[attr_name], path[1:])
         elif attr_name in dir(obj):
@@ -380,18 +419,30 @@ def _get_attr(obj, path):
         File = get_module("mo_files").File
         possible_error = None
         python_file = (File(obj.__file__).parent / attr_name).set_extension("py")
-        python_module = (File(obj.__file__).parent / attr_name / "__init__.py")
+        python_module = File(obj.__file__).parent / attr_name / "__init__.py"
         if python_file.exists or python_module.exists:
             try:
                 # THIS CASE IS WHEN THE __init__.py DOES NOT IMPORT THE SUBDIR FILE
                 # WE CAN STILL PUT THE PATH TO THE FILE IN THE from CLAUSE
                 if len(path) == 1:
                     # GET MODULE OBJECT
-                    output = __import__(obj.__name__ + str(".") + str(attr_name), globals(), locals(), [str(attr_name)], 0)
+                    output = __import__(
+                        obj.__name__ + str(".") + str(attr_name),
+                        globals(),
+                        locals(),
+                        [str(attr_name)],
+                        0,
+                    )
                     return output
                 else:
                     # GET VARIABLE IN MODULE
-                    output = __import__(obj.__name__ + str(".") + str(attr_name), globals(), locals(), [str(path[1])], 0)
+                    output = __import__(
+                        obj.__name__ + str(".") + str(attr_name),
+                        globals(),
+                        locals(),
+                        [str(path[1])],
+                        0,
+                    )
                     return _get_attr(output, path[1:])
             except Exception as e:
                 Except = get_module("mo_logs.exceptions.Except")
@@ -400,7 +451,11 @@ def _get_attr(obj, path):
         # TRY A CASE-INSENSITIVE MATCH
         matched_attr_name = lower_match(attr_name, dir(obj))
         if not matched_attr_name:
-            get_logger().warning(PATH_NOT_FOUND + "({{name|quote}}) Returning None.", name=attr_name, cause=possible_error)
+            get_logger().warning(
+                PATH_NOT_FOUND + "({{name|quote}}) Returning None.",
+                name=attr_name,
+                cause=possible_error,
+            )
         elif len(matched_attr_name) > 1:
             get_logger().error(AMBIGUOUS_PATH_FOUND + " {{paths}}", paths=attr_name)
         else:
@@ -427,10 +482,12 @@ def _get_attr(obj, path):
 
 def _set_attr(obj_, path, value):
     obj = _get_attr(obj_, path[:-1])
-    if obj is None:  # DELIBERATE USE OF `is`: WE DO NOT WHAT TO CATCH Null HERE (THEY CAN BE SET)
+    if (
+        obj is None
+    ):  # DELIBERATE USE OF `is`: WE DO NOT WHAT TO CATCH Null HERE (THEY CAN BE SET)
         obj = _get_attr(obj_, path[:-1])
         if obj is None:
-            get_logger().error(PATH_NOT_FOUND+" tried to get attribute of None")
+            get_logger().error(PATH_NOT_FOUND + " tried to get attribute of None")
 
     attr_name = path[-1]
 
@@ -444,7 +501,9 @@ def _set_attr(obj_, path, value):
         elif value == None:
             new_value = None
         else:
-            new_value = _get(old_value, CLASS)(value)  # TRY TO MAKE INSTANCE OF SAME CLASS
+            new_value = _get(
+                old_value, CLASS
+            )(value)  # TRY TO MAKE INSTANCE OF SAME CLASS
     except Exception:
         old_value = None
         new_value = value
@@ -464,7 +523,27 @@ def lower_match(value, candidates):
     return [v for v in candidates if v.lower() == value.lower()]
 
 
-def wrap(v):
+def dict_to_data(d):
+    """
+    DO NOT CHECK TYPE
+    :param d: dict
+    :return: Data
+    """
+    m = _new(Data)
+    _set(m, SLOT, d)
+    return m
+
+
+def list_to_data(v):
+    """
+    to_data, BUT WITHOUT CHECKS
+    """
+    output = _new(FlatList)
+    _set(output, LIST, v)
+    return output
+
+
+def to_data(v):
     """
     WRAP AS Data OBJECT FOR DATA PROCESSING: https://github.com/klahnakoski/mo-dots/tree/dev/docs
     :param v:  THE VALUE TO WRAP
@@ -474,27 +553,36 @@ def wrap(v):
     type_ = _get(v, CLASS)
 
     if type_ in (dict, OrderedDict):
-        m = object.__new__(Data)
+        m = _new(Data)
         _set(m, SLOT, v)
         return m
     elif type_ is none_type:
         return Null
     elif type_ is list:
-        return FlatList(v)
+        return list_to_data(v)
     elif type_ in generator_types:
-        return FlatList(list(unwrap(vv) for vv in v))
+        return list_to_data(list(from_data(vv) for vv in v))
     else:
         return v
 
 
-def wrap_leaves(value):
+wrap = to_data
+
+
+def leaves_to_data(value):
     """
     dict WITH DOTS IN KEYS IS INTERPRETED AS A PATH
     """
-    return wrap(_wrap_leaves(value))
+    return to_data(_leaves_to_data(value))
 
 
-def _wrap_leaves(value):
+wrap_leaves = leaves_to_data
+
+
+def _leaves_to_data(value):
+    """
+    RETURN UNWRAPPED STRUCTURES
+    """
     if value == None:
         return None
 
@@ -503,11 +591,11 @@ def _wrap_leaves(value):
         return value
     if class_ in data_types:
         if class_ is Data:
-            value = unwrap(value)
+            value = from_data(value)
 
         output = {}
         for key, value in value.items():
-            value = _wrap_leaves(value)
+            value = _leaves_to_data(value)
 
             if key == "":
                 get_logger().error("key is empty string.  Probably a bad idea")
@@ -533,16 +621,16 @@ def _wrap_leaves(value):
                 else:
                     d[seq[-1]] = value
         return output
-    if hasattr(value, '__iter__'):
+    if hasattr(value, "__iter__"):
         output = []
         for v in value:
-            v = wrap_leaves(v)
+            v = leaves_to_data(v)
             output.append(v)
         return output
     return value
 
 
-def unwrap(v):
+def from_data(v):
     if v is None:
         return None
     _type = _get(v, CLASS)
@@ -560,11 +648,14 @@ def unwrap(v):
         else:
             return v
     elif _type in generator_types:
-        return (unwrap(vv) for vv in v)
+        return (from_data(vv) for vv in v)
     else:
         return v
 
 
+unwrap = from_data
+
+
 def listwrap(value):
     """
     PERFORMS THE FOLLOWING TRANSLATION
@@ -596,11 +687,15 @@ def listwrap(value):
     if value == None:
         return FlatList()
     elif is_list(value):
-        return wrap(value)
+        if isinstance(value, list):
+            return list_to_data(value)
+        else:
+            return value
     elif is_many(value):
-        return wrap(list(value))
+        return list_to_data(list(value))
     else:
-        return wrap([unwrap(value)])
+        return list_to_data([from_data(value)])
+
 
 def unwraplist(v):
     """
@@ -610,11 +705,11 @@ def unwraplist(v):
         if len(v) == 0:
             return None
         elif len(v) == 1:
-            return unwrap(v[0])
+            return from_data(v[0])
         else:
-            return unwrap(v)
+            return from_data(v)
     else:
-        return unwrap(v)
+        return from_data(v)
 
 
 def tuplewrap(value):
@@ -623,16 +718,52 @@ def tuplewrap(value):
     """
     if is_many(value):
         return tuple(tuplewrap(v) if is_sequence(v) else v for v in value)
-    return unwrap(value),
+    return (from_data(value),)
 
 
-from mo_dots.datas import Data, SLOT, data_types, is_data
-from mo_dots.nones import Null, NullType
-from mo_dots.lists import FlatList, is_list, is_sequence, is_container, is_many
-from mo_dots.objects import DataObject
+def is_null(t):
+    # RETURN True IF EFFECTIVELY NOTHING
+    class_ = t.__class__
+    if class_ in (none_type, NullType):
+        return True
+    else:
+        try:
+            return t == None
+        except Exception:
+            return False
+
+
+def is_missing(t):
+    # RETURN True IF EFFECTIVELY NOTHING
+    class_ = t.__class__
+    if class_ in (none_type, NullType):
+        return True
+    elif class_ in data_types:
+        return False
+    elif class_ in finite_types and not t:
+        return True
+    else:
+        return t == None
+
 
 # EXPORT
-import mo_dots.nones as temp
-temp.wrap = wrap
-temp.is_sequence = is_sequence
-del temp
+export("mo_dots.nones", to_data)
+
+export("mo_dots.datas", to_data)
+export("mo_dots.datas", from_data)
+export("mo_dots.datas", coalesce)
+export("mo_dots.datas", _getdefault)
+export("mo_dots.datas", hash_value)
+export("mo_dots.datas", listwrap)
+export("mo_dots.datas", literal_field)
+
+export("mo_dots.lists", list_to_data)
+export("mo_dots.lists", to_data)
+export("mo_dots.lists", coalesce)
+export("mo_dots.lists", from_data)
+
+export("mo_dots.objects", list_to_data)
+export("mo_dots.objects", to_data)
+export("mo_dots.objects", from_data)
+export("mo_dots.objects", get_attr)
+export("mo_dots.objects", set_attr)
