@@ -10,22 +10,14 @@
 from __future__ import absolute_import, division, unicode_literals
 
 import importlib
-import sys
 
-from mo_future import PY2, text
+from mo_future import STDOUT, STDERR
 
-OBJ = text("_obj")
-CLASS = text("__class__")
+OBJ = str("_obj")
+CLASS = str("__class__")
 
 _Log = None
 
-if PY2:
-    STDOUT = sys.stdout
-    STDERR = sys.stderr
-else:
-    STDOUT = sys.stdout.buffer
-    STDERR = sys.stderr.buffer
-
 
 def get_logger():
     global _Log
@@ -40,7 +32,6 @@ def get_logger():
         return _Log
 
 
-
 def get_module(name):
     try:
         return importlib.import_module(name)
